__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import citelang.main.packages as packages
from citelang.logger import logger


def get_package(manager, name, version=None, client=None, data=None, use_cache=True):
    """
    Return a package handle for a custom package (defined in citelang) or libraries.io
    """
    # Case 1: a custom manager
    if manager in packages.manager_names:
        return CustomPackage(
            name=name,
            manager=manager,
            version=version,
            client=client,
            data=data,
            use_cache=use_cache,
        )

    # Case 2: libraries.io
    return Package(
        name=name,
        manager=manager,
        version=version,
        client=client,
        data=data,
        use_cache=use_cache,
    )


class PackageBase:
    def __init__(
        self, manager, name, version=None, client=None, data=None, use_cache=True
    ):
        """
        The version can be set explicitly or come from the name.
        """
        self.manager = manager
        self.version = version
        self.parse(name)
        self.data = data or {}
        self.latest = None
        self.set_client(client)
        self.use_cache = use_cache

    def __repr__(self):
        return str(self)

    def __str__(self):
        return "[citelang-package]"

    @property
    def cache_name(self):
        return f"package/{self.manager}/{self.name}"

    def set_client(self, client=None):
        """
        A package needs a client controller to make continued calls.
        """
        if not client:
            from .client import Client

            self.client = Client()
        self.client = client

    def parse(self, name):
        """
        Parse a manager and name into the package
        """
        if "@" in name:
            name, version = name.split("@", 1)
            self.version = version
        if "[" in name and "]" in name:
            name = name.split("[", 1)[0]

        # invalid characters found!
        name = name.replace(";", "")
        self.name = name

    def dependencies(self, return_data=False):
        raise NotImplementedError

    def info(self):
        raise NotImplementedError


class CustomPackage(PackageBase):
    """
    A wrapper for a custom package (provided by citelang)
    """

    def dependencies(self, return_data=False):
        """
        Retrieve custom package dependencies
        """
        manager = packages.managers[self.manager]()

        # Only will be used if we have a version
        cache_name = f"dependencies/{self.manager}/{self.name}/{self.version}"

        # If we don't have a version, we have to retrieve an update
        if not self.version:
            deps = manager.package(self.name)
            result = self.client.get_endpoint("dependencies", data=deps)
            if return_data:
                return result.data.get("dependencies", [])
            return result

        # Case 2: Try to use the cache
        result = self.client.get_cache(cache_name)
        if not result or not self.use_cache:
            version = "@%s" % self.version if self.version else ""
            logger.info("Retrieving new result for %s@%s..." % (self.name, version))
            deps = manager.package(self.name)
            result = self.client.get_endpoint("dependencies", data=deps)
        else:
            result = self.client.get_endpoint("dependencies", data=result)

        # If cache is enabled, we save the result
        self.client.cache(cache_name, result)

        # Return the wrapped result or raw data
        if return_data:
            return result.data.get("dependencies", [])
        return result

    def info(self):
        """
        Get info for a libraries.io or custom package
        """
        manager = packages.managers[self.manager]()

        # First try retrieving from the cache
        result = self.client.get_cache(self.cache_name)
        if not result or not self.use_cache:
            version = "@%s" % self.version if self.version else ""
            logger.info(
                "Retrieving new result for package %s@%s..." % (self.name, version)
            )
            pkg = manager.package(self.name)
            result = self.client.get_endpoint(
                "package", data=pkg, package_name=self.name, manager=self.manager
            )
        else:
            result = self.client.get_endpoint(
                "package", data=result, package_name=self.name, manager=self.manager
            )

        # If cache is enabled, we save the result
        self.client.cache(self.cache_name, result)
        return result


class Package(PackageBase):
    """
    A basic wrapper for package parsing functions (libraries.io)
    """

    def dependencies(self, return_data=False):
        """
        Retrieve libraries.io dependencies
        """
        if not self.version:
            self.info()
            self.version = self.latest

        # Check for result in cache
        cache_name = f"dependencies/{self.manager}/{self.name}/{self.version}"
        result = self.client.get_cache(cache_name)
        if not result or not self.use_cache:
            logger.info(
                "Retrieving new result for %s@%s dependencies..."
                % (self.name, self.version)
            )
            result = self.client.get_endpoint(
                "dependencies",
                manager=self.manager,
                package_name=self.name,
                version=self.version,
            )
        else:
            result = self.client.get_endpoint("dependencies", data=result)

        # If cache is enabled, we save the result
        self.client.cache(cache_name, result)
        if return_data:
            return result.data.get("dependencies", [])
        return result

    def info(self):
        """
        Get info for a libraries.io package
        """
        result = self.client.get_cache(self.cache_name)
        if not result or not self.use_cache:
            version = "@%s" % self.version if self.version else ""
            logger.info(
                "Retrieving new result for package %s@%s..." % (self.name, version)
            )
            result = self.client.get_endpoint(
                "package",
                manager=self.manager,
                package_name=self.name,
            )
        else:
            result = self.client.get_endpoint(
                "package", data=result, manager=self.manager, package_name=self.name
            )

        # Set a latest version if we find one
        if "versions" in result.data and result.data["versions"]:
            self.latest = result.data["versions"][-1]["number"]
        return result



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

from citelang.logger import logger
import citelang.utils as utils
import citelang.main.graph as graph

from rich.table import Table as RichTable
from rich.console import Console

import random
import os


class Result:
    """
    A result holds the request result and can parse or return in different formats.
    """

    def __init__(self, data, endpoint):
        self.endpoint = endpoint

        # Does the endpoint want to sort or otherwise order?
        data = data or {}
        self.data = self.endpoint.order(data)

    def print_json(self):
        print(utils.print_json(self.data))

    def to_dict(self):
        return self.data

    def save(self, outfile):
        """
        Save to output file
        """
        logger.info("Saving to %s..." % outfile)
        utils.write_json(self.data, outfile)


class Table(Result):
    """
    A table is a result formatted as a table for the client.
    """

    def __init__(self, data, endpoint):
        super().__init__(data, endpoint)

        # Keep track of the max length for each field not truncated
        self.max_widths = {}
        self.ensure_complete()

    def available_width(self, columns):
        """
        Calculate available width based on fields we cannot truncate (urls)
        """
        # We will determine column width based on terminal size
        try:
            width = os.get_terminal_size().columns
        except OSError:
            width = 120

        # Calculate column width
        column_width = int(width / len(columns))
        updated = width

        for _, needed in self.max_widths.items():
            updated = updated - needed

        # We don't have enough space
        if updated < 0:
            logger.warning("Terminal is too small to correctly render!")
            return column_width

        # Otherwise, recalculate column width taking into account truncation
        # We use the updated smaller width, and break it up between columns
        # that don't have a max width
        return int(updated / (len(columns) - len(self.max_widths)))

    def ensure_complete(self):
        """
        If any data missing fields, ensure they are included
        """
        if isinstance(self.data, list):
            self.ensure_complete_list()
        # We don't check other types for now

    def ensure_complete_list(self):
        """
        Given a list result, check the fields.
        """
        fields = set()
        for entry in self.data:
            [fields.add(x) for x in entry.keys()]

        # Ensure fields are present
        for entry in self.data:
            for field in fields:

                if field not in entry:
                    entry[field] = ""

                # We can't truncate this field!
                if field in self.endpoint.truncate_list:

                    # We haven't seen it before
                    if field not in self.max_widths:
                        self.max_widths[field] = len(entry[field])
                    elif self.max_widths[field] < len(entry[field]):
                        self.max_widths[field] = len(entry[field])

    @property
    def color(self):
        """
        Return a random color
        """
        return "color(" + str(random.choice(range(255))) + ")"

    def table_columns(self):
        """
        Shared function to return consistent table columns
        """
        # An endpoint that returns structured json can choose to flatten
        data = self.endpoint.table_data(self.data)

        # Get column titles
        columns = []
        contenders = list(data[0].keys())
        for column in contenders:
            if column in self.endpoint.skip_list:
                continue
            columns.append(column)
        return columns

    def table_rows(self, columns, limit=25):
        """
        Shared function to yield rows as a list
        """
        # All keys are lowercase
        column_width = self.available_width(columns)

        # An endpoint that returns structured json can choose to flatten
        data = self.endpoint.table_data(self.data)

        for i, row in enumerate(data):

            # have we gone over the limit?
            if limit and i > limit:
                return

            parsed = []
            for column in columns:
                content = str(row[column]) if row[column] else ""
                if (
                    content
                    and len(content) > column_width
                    and column not in self.endpoint.truncate_list
                ):
                    content = content[:column_width] + "..."
                parsed.append(content)
            yield parsed

    def table(self, limit=25):
        """
        Pretty print a table of results.
        """
        table = RichTable(title=self.endpoint.title)

        # Don't reuse colors
        seen_colors = []

        # No dependencies!
        data = self.endpoint.table_data(self.data)
        if not data:
            print("This package does not have any dependencies.")
            return

        # Get column titles
        columns = self.table_columns()
        for column in columns:
            color = None
            while not color or color in seen_colors:
                color = self.color
            title = " ".join([x.capitalize() for x in column.split("_")])
            table.add_column(title, style=color)

        # Add rows
        for row in self.table_rows(columns, limit=limit):
            table.add_row(*row)

        # And print!
        console = Console()
        console.print(table, justify="center")


class Tree(Result):
    def __init__(self, result):
        self.result = result
        self.parse_data()

    def parse_data(self):
        """
        Parse result into data
        """
        data = {
            "name": self.result.name,
            "credit": self.result.credit,
            "weight": self.result.weight,
            "round_by": self.result.round_by,
            "children": [],
        }

        def add(data, children, total):
            total += data.credit
            for child in data.children:
                node = {
                    "name": child.name,
                    "credit": child.credit,
                    "weight": child.weight,
                    "round_by": child.round_by,
                    "children": [],
                }
                total = add(child, node["children"], total)
                children.append(node)
            return total

        total = add(self.result, data["children"], 0)
        self.data = data
        self.data["total"] = total

    def print_result(self):
        """
        Print a tree result (todo redo with color and rich)
        """

        def print_result(result, indent=2):
            print(
                "%s%20s: %s"
                % (
                    " " * indent,
                    result["name"],
                    round(result["credit"], result["round_by"]),
                )
            )
            for child in result["children"]:
                print_result(child, indent=indent * 2)

        print_result(self.data)
        print("total: %s" % round(self.data["total"], 3))


class Graph(Result):
    """
    A graph result can generate text for a graph
    """

    def __init__(self, root):
        self.root = root

    def graph(self, fmt=None):
        """
        Generate a graph of dependencies
        """
        # Select output format (default to console)
        if fmt == "dot":
            out = graph.Dot(self.root)
        elif fmt == "cypher":
            out = graph.Cypher(self.root)
        elif fmt == "gexf":
            out = graph.Gexf(self.root)
        else:
            # Only console supported for now!
            out = graph.Console(self.root)
        return out.generate()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2022, Vanessa Sochat"
__license__ = "MPL 2.0"

import citelang.main.base as base
import citelang.main.graph as graph
import citelang.main.result as results
import citelang.main.package as package
import citelang.main.packages as packages
from citelang.logger import logger


class Client(base.BaseClient):
    """
    Interact with a libaries.io or custom package server via CiteLang
    """

    def dependencies(self, manager, name, use_cache=True):
        """
        Get dependencies for a package. If no version, use latest.
        """
        self.check_manager(manager, use_cache)
        pkg = package.Package(manager, name, client=self, use_cache=use_cache)
        return pkg.dependencies()

    def package_managers(self, use_cache=True):
        """
        Get a listing of package managers supported on libraries.io
        """
        result = self.get_cache("package_managers")
        if not result or not use_cache:
            logger.info("Retrieving new result for package managers...")
            result = self.get_endpoint("package_managers")

            # Update custom package managers
            result.data += [p().info() for _, p in packages.managers.items()]
        else:
            result = self.get_endpoint("package_managers", data=result)

        # If cache is enabled, we save the result
        self.cache("package_managers", result)
        return result

    def check_manager(self, name, use_cache=True):
        """
        Check if a package manager is supported
        """
        if name in packages.manager_names:
            return
        managers = self.package_managers(use_cache)

        # Ensure the manager is allowed
        if managers.data:
            managers = [x["name"].lower() for x in managers.data]
            if name not in managers:
                managers = "\n".join(managers)
                logger.exit(f"{name} is not a known manager. Choices are:\n{managers}")

    def graph(
        self,
        manager,
        name,
        use_cache=True,
        max_depth=None,
        max_deps=None,
        min_credit=0.01,
        credit_split=0.5,
        fmt=None,
    ):
        """
        Generate a graph for a package.

        credit_split is how to split credit between some package and its dependents. E.g., 0.5 means 50/50. 0.8 means
        the main package gets 80%, and dependencies split 20%. We go up until the min credit 0.05 at which case we
        stop adding.
        """
        root = self._graph(
            manager=manager,
            name=name,
            use_cache=use_cache,
            max_depth=max_depth,
            max_deps=max_deps,
            min_credit=min_credit,
            credit_split=credit_split,
        )
        return results.Graph(root).graph(fmt=fmt)

    def _graph(
        self,
        manager,
        name,
        use_cache=True,
        max_depth=None,
        max_deps=None,
        min_credit=0.01,
        credit_split=0.01,
    ):
        """
        Shared 'private' function to generate graph
        """
        self.check_manager(manager, use_cache)
        pkg = package.get_package(manager, name, client=self, use_cache=use_cache)

        # keep track of deps (we only care about name, not version)
        seen = set()

        # Top node gets full credit 1.0 (to split between itself and deps)
        root = graph.Node(
            obj=pkg,
            weight=1.0,
            credit_split=credit_split,
            depth=0,
            min_credit=min_credit,
        )

        # A pointer to the next node
        next_nodes = [root]

        # Booleans to trigger exiting parser
        stop_looking = False

        # Keep handle to previous children in case we stop looking
        previous = []

        while next_nodes and not stop_looking:
            next_node = next_nodes.pop(0)
            deps = next_node.obj.dependencies(return_data=True)

            # Stopping point - exceeded max depth
            if max_depth and next_node.depth > max_depth:
                stop_looking = True

            # Stopping point - exceeded max deps (plus 1 for original package)
            if max_deps and len(seen) + 1 > max_deps:
                stop_looking = True

            seen.add(next_node.name)

            if not deps:
                continue

            # How much credit for each of the new deps?
            # credit for all deps AS a percentage of the weight BROKEN into number of deps
            dep_credit = ((1 - credit_split) * next_node.weight) / len(deps)

            # Stopping point - dependency credit is too small
            if dep_credit < min_credit:
                stop_looking = True

            # If we are stopping, time to break and distribute remaining credit
            # to nodes we couldn't parse children for.
            if stop_looking:

                # We haven't parsed this one yet
                next_node.total_credit = next_node.weight

                # Give all remaining credit to the node we aren't parsing
                for node in previous:

                    # We might have added and given credit to some children
                    if node.children:

                        # Redistribute credit amongst children that were > threshold
                        dep_credit = ((1 - credit_split) * node.weight) / len(
                            node.children
                        )
                        for child in node.children:
                            child.total_credit = dep_credit

                        # The node's credit is that weight minus total dep credit
                        node.total_credit = node.weight - (
                            dep_credit * len(node.children)
                        )
                    else:
                        node.total_credit = node.weight
                break

            # Calculate credit for each dependency and add as child
            for dep in deps:

                # Haven't seen this case, but just a check
                dep_name = dep["name"] or dep["project_name"]
                if not dep_name:
                    continue
                depnode = package.get_package(
                    manager, dep_name, client=self, use_cache=use_cache
                )
                child = graph.Node(
                    obj=depnode,
                    weight=dep_credit,
                    credit_split=credit_split,
                    depth=next_node.depth + 1,
                    min_credit=min_credit,
                )
                next_node.add_child(child)
                next_nodes.append(child)

                # Store previous if we need to update weight
                previous.append(child)

        return root

    def credit(
        self,
        manager,
        name,
        use_cache=True,
        max_depth=None,
        max_deps=None,
        min_credit=0.05,
        credit_split=0.5,
    ):
        """
        Get the credit root node, then do additional graph parsing.
        """
        root = self._graph(
            manager=manager,
            name=name,
            use_cache=use_cache,
            max_depth=max_depth,
            max_deps=max_deps,
            min_credit=min_credit,
            credit_split=credit_split,
        )
        return results.Tree(root)

    def package(self, manager, name, use_cache=True):
        """
        Lookup a package in a specific package manager
        """
        self.check_manager(manager, use_cache)
        pkg = package.get_package(manager, name, client=self, use_cache=use_cache)
        return pkg.info()



from .terminal import (
    run_command,
    get_installdir,
    confirm_action,
    confirm_uninstall,
)
from .fileio import (
    get_tmpfile,
    get_tmpdir,
    mkdir_p,
    print_json,
    read_file,
    read_json,
    write_file,
    write_json,
)
